%
%                       Munorm Version 1.2
%
%         Craig T. Lawrence,*  Andre L. Tits* and Vasile Sima+
%
%   *Institute for Systems Research and Department of Electrical Engineering
%                      University of Maryland
%                      College Park, MD 20742
%                             USA
%
%   +National Institute for Research & Development in Informatics
%                      Bucharest, Romania
%
%                          January 2003
%
%  Copyright (c) 1995-2003 by C.T. Lawrence, A.L. Tits and V. Sima. 
%  All rights reserved.
%
% This function computes the supremum over frequency of the standard "D-G"
% upper bound to the structured singular value of a finite dimensional system
% given a state-space realization (A,B,C,E).  With some minor variations,
% it implements the  algorithm proposed in [LTV].  It handles the
% general "mixed-mu" problem in both continuous time and discrete time.
%
%
% Synopsis:
%    [mu_nm,w_hat,iwarn,iter] = munorm(A,B,C,E,blk,options,tol)
% Input arguments:
%    A,B,C,E: state-space quadruple
%         Note: it is assumed that the transfer function is square.
%    blk: block structure; same syntax as in mutools, i.e.,
%            nblk x 2, where nblk is the number of blocks in the
%            structure; the ith row [i1 i2] corresponds to the
%            ith block.  If i1>0, the block is complex, if i1<0,
%            it is real.  If i2=0, the block is of the "repeated"
%            type (scalar times identity) of size |i1| x |i1|.
%            If i2>0, the block is "full", of size i1 x i2, and
%            i1 must be positive.
%    The last two arguments are optional:
%    options(1) = {0,1}, 0 = continuous time, 1 = discrete time.  Default = 0.
%    options(2) = {0,1,2}, method for choosing next trial frequency
%      0 = always use the midpoint rule, 
%      1 = attempt cubic interpolation when Omega consists of a single interval
%      2 = attempt rational interpolation when Omega consists of a single
%           interval.
%      Default = 2.
%    options(3) = {0,1,2}, option for printing the (intermediate) results
%      0 = print nothing, 
%      1 = print iteration number and the current muhat value,
%      2 = print iteration number, the current best value muhat of the munorm
%          estimate, the corresponding frequency w_hat, and the stop reason.
%      Default = 0.
%    tol: tolerance: the algorithm terminates when either the computed
%      value is guaranteed to be within tol of the true supremum, or
%      the last two trial frequency points were within tol of each other.
%      Default = 1.e-4.
% Output arguments
%    mu_nm: supremum over frequency of the mu upper bound.
%    w_hat: frequency at which the supremum is approximately achieved.
%    iwarn: optional warning indicating the stop reason:
%      0 = no warning, 
%      1 = no intervals were found,
%      2 = empty intersection was found.
%    iter: optional argument returning the number of iterations performed.
%  
% Reference:
% [LTV] C.T. Lawrence, A.L. Tits, P. Van Dooren, A Fast Algorithm for the 
% Computation of an Upper Bound on the Mu-Norm, Automatica, v. 36, n. 3, 
% pp. 449-456, 2000.  (Also available from the authors' home pages.)
%
%
%                   Conditions for External Use
%                   ===========================
%
%   1. This code  may not be  distributed to third parties. Interested
%      parties  should contact  the authors directly.
%   2. If modifications are performed on the code, these modifications
%      shall be  communicated to  the authors.  The modified  routines
%      will  remain the sole  property of the authors.
%   3. Due  acknowledgment  must  be  made of the  use of this code in
%      research  reports  or  publications.  Whenever such reports are
%      released for  public  access, a copy should be forwarded to the
%      authors.
%   4. This code may only be used for research and development, unless
%      it has been agreed  otherwise with the authors in writing.
%
%
%           Enhancements in successive versions of MUNORM.m
%           ================================================
%
%   Version 1.2 (January 2003):
%       - Added feature: displaying the successive iterates is now
%         an option instead of being forced.
%       - Added feature: the reason of premature stop (if any) and the 
%         number of iterations are now optionally returned in the
%         output arguments iwarn and iter, respectively.
%       - Bug fix: In Version 1.1b, the erroneous computation of Gk
%         in Step 1 had been fixed at one place, but not in a second
%         place; this is now done.
%       - Bug fix: the expression for dEF was incorrect; it is now fixed.
%       - An error exit now occurs if input argument tol is negative.
%       - Bug fix: all imaginary eigenvalues (of the Hamiltonian matrix) 
%         with non-negative, rather than positive, imaginary part are now
%         used in the determination of "intervals".  This prevents faulty
%         behavior on certain problems.
%       - Bug fix: fine-tuned the tests to detect imaginary eigenvalues of 
%         the Hamiltonian matrix, and to correctly count imaginary eigenvalues
%         of very small magnitude.
%       - Bug fix: fixed a problem (in the processing of intervals) that 
%         sometimes unduly caused termination at iteration 1.
%   Version 1.1b (August 2002):
%       - Bug fix: wtil was not cleared after each iteration nor clipped
%         to the new dimension, so old values were sometimes erroneously
%         included.  Changed [wtil,I]=sort(wtil) to [wtil,I]=sort(wtil(1:nw)).
%         (Thanks go to J.M. Biannic for producing an example that
%         uncovered the bug.)
%       - Bug fix: Computation of Gk in Step 1 was incorrect.
%       - Bug fix: Computation of EF was incorrect.
%       - Bug fix: Error in the interval determination logic in the case
%         of a negative slope at 0.
%       - Semi-bug fix: use exact, rather than perturbed, Hamiltonian
%         matrix.  (Unclear why we used to perturb it, but it cause
%         problems on many examples.)
%       - Semi-bug fix: terminates if the set of intervals becomes empty.
%       - Fine tuning: Measure for intervals was changed from atan to 
%         scaled atan.
%       - Fine tuning: the tolerance, in assessing whether any of the new
%         w point is the old wk, was increased, due to apparent lack of
%         accuracy in the computation of eigenvalues of the Hamiltonian
%         matrix.
%       - Fine tuning: reduced the threshold for that test that triggers
%         checking the stopping criterion.
%   Version 1.1a (October 2001):
%       - Change    [A,B,C,E] = [Ai,Bi,Ci,Ei];  to    A=Ai; B=Bi; C=Ci; E=Ei;
%         (the former now gives a compile error);
%       - Display appropriate text before the pause.
%       - Return "w_hat" in addition to "mu_nm".  (Suggested by J.M. Biannic.)
%         (Also renamed "what" to "w_hat".)
%       - Bug fix: when computing interval "lengths" to determine the
%         largest interval, "tan" was erroneously used instead of "atan"
%         (four instances in two lines of code).  (Thanks go to J.M. Biannic
%         for producing an example that uncovered the bug.)
%   Version 1.1 (September 1999): Fix a number of bugs in the mixed-mu case,
%       allow for infinite intervals (in the case of not-strictly-proper 
%       systems), handle discrete-time systems.
%   Version 1.0: June 1995.  Mostly for complex-mu, continuous time,
%       strictly proper systems.  This version was not publicly released.
%
%
function [mu_nm,w_hat,iwarn,iter] = munorm(Ai,Bi,Ci,Ei,blk,options,tol)

% Initialization
FALSE=0; TRUE=1;
discrete=FALSE;         % TRUE for discrete-time system
tol1 = 1.e-4;           % default stopping tolerance
j=sqrt(-1);
interp=2;               % 0=midpoint only, 1=cubic, 2=rational
info = 0;
if nargin >= 6
   if length(options) > 2, info = options(3); end
   if options(1) == 1
      discrete = TRUE;
   end
   interp = options(2);
end
if nargin == 7
   tol1 = tol;
   if (tol<0),
       error('Data error: tol must be nonnegative');
   end
end

A=Ai; B=Bi; C=Ci; E=Ei;
if discrete == TRUE
   % Apply bilinear transform to convert to continuous time
   Ad = A; Bd = B; Cd = C; Ed = E;
   [A,B,C,E] = bilinear(-Ad,Bd,-Cd,Ed,0.5);
end
eigA = eig(A);

[n,m]=size(B);            % Problem size
iter=1;                   % Iteration counter
wk=0;                     % Trial frequency
wksaved=-1;               % Previous trial frequency
muhat=0;                  % Best value so far
w_hat=wk;                 % Best frequency so far
wc=-1;                    % Trial frequency obtained by interpolation
mu_k_old=-1;              % Previous estimate of munorm
done = FALSE;             % Flag indicating some stopping criterion satisfied
check_done = FALSE;       % Flag indicating we should check stopping criterion
check_flat = FALSE;       % Flag for checking if mu = constant
intervals=[];             % List of intervals where maximizer may lie
backup=[-1 0 0; -1 0 0];  % Store information for interpolation
backup_ptr=1;
nf=0;                     % Number of frequencies corresponding to sigma_max
nint=1;                   % Number of intervals
iwarn=0;                  % Warning indicator

% Check to see if the problem is mixed-mu, or pure complex
mixed = FALSE;            % Flag indicating mixed mu
[nb, temp] = size(blk);
for index=1:nb,
   if blk(index,1) < 0
      mixed = TRUE;
      break;
   end
end

% Main loop of algorithm
while done == FALSE,
   record_slope=FALSE;    % Flag indicating whether data from current
           			      % trial point should be used for interpolation

   % Step 1 - Compute Dk, Gk, mu_k, and muhat 

   if (check_done==FALSE)
      mat= C*inv(j*wk*eye(n) - A)*B + E;
      if mixed == TRUE
         [bnds,dvec,sens,pvec,gvec] = mu(mat,blk,'uCU');
         mu_k=bnds(1,1);
         [dl,dr,gl,gm,gr]=muunwrap(dvec,gvec,blk);
         % numbering below is as in P.M. Young' Ph.D. 
         % dissertation, pp. 104-105.
         d4=dl; g4=gl;
         gtmp = inv(sqrtm(eye(size(g4))+g4^2));
         d1 = d4 * gtmp * d4;
         g1 = mu_k .* d4 * g4 * gtmp * d4;
         d2 = sqrtm(d1);
         g2 = inv(sqrtm(d1)) * g1 * inv(sqrtm(d1));
         d3 = d2;
         g3 = (1/mu_k) .* g2;
         Dk = d3;
         Gk = g3;
      else 
         [bnds,dvec,sens,pvec] = mu(mat,blk,'uCU');
         mu_k=bnds(1,1);
    	 [Dk, Dr] = unwrapd(dvec,blk);
         Gk = zeros(m);
      end

      % check to see if interpolation gives better value
      % than the mid-point rule
      if (wc >= 0)
         mat= C*inv(j*wc*eye(n) - A)*B + E;
         if mixed == TRUE
            [bnds,dvec,sens,pvec,gvec] = mu(mat,blk,'uCU');
            mu_kc=bnds(1,1);
            [dl,dr,gl,gm,gr]=muunwrap(dvec,gvec,blk);
            d4=dl; g4=gl;
            gtmp = inv(sqrtm(eye(size(g4))+g4^2));
            d1 = d4 * gtmp * d4;
            g1 = mu_kc .* d4 * g4 * gtmp * d4;
            d2 = sqrtm(d1);
            g2 = inv(sqrtm(d1)) * g1 * inv(sqrtm(d1));
            d3 = d2;
            g3 = (1/mu_kc) .* g2;
            Dkc = d3;
            Gkc = g3;
         else
            [bnds,dvec,sens,pvec] = mu(mat,blk,'uCU');
            mu_kc=bnds(1,1);
	        [Dkc, Dr] = unwrapd(dvec,blk);
            Gkc = zeros(m);
         end
         if mu_kc>mu_k
	        % Interpolation point was better, accept
            mu_k=mu_kc;
            Dk=Dkc;
            Gk=Gkc;
            wk=wc;
         end
      end

      % Should we check the main stopping criterion this time?
      if (abs(mu_k-mu_k_old) < tol1)   
         check_done=TRUE;
      end

      % Check if trial frequency essentially unchanged
      if wksaved > -1
         if (abs(wk - wksaved) < tol1)
            done = TRUE;
            break;
         end
      end
      wksaved = wk;

      % Check for flat mu-curve
      if (check_flat == TRUE)
         if (abs(mu_k-mu_k_old) < 10*tol1) 
	        done = TRUE;
	        break;
	     end
    	 check_flat = FALSE;
      end
      if (mu_k_old < 0)
   	     check_flat=TRUE;
      end

      if mu_k > muhat
         muhat=mu_k;
         w_hat=wk;
         record_slope=TRUE;
      end
   else
      check_done=FALSE;
   end

   % Step 2 - Compute collection of intervals where maximizer
   %          could lie.

   % Form Hamiltonian and get eigenvalues
   xi = muhat;
   if (check_done == TRUE)
      xi = xi + tol1;
   end
   temp0 = inv(sqrtm(eye(m) + Gk*Gk));
   Dinv = inv(Dk);
   BF = (1/sqrt(xi))*B*Dinv*temp0;
   CF = (1/sqrt(xi))*Dk*C;
   EF = ((1/xi)*(Dk*E*Dinv) - j*Gk)*temp0;
   temp1 = [BF zeros(n, m); zeros(n, m) -CF'];
   %     H(1,A,BF,CF,EF) is replaced with H(1.001,A,BF,CF,EF) to
   %     circumvent numerical difficulties when Gk is large
   temp2 = inv([-EF eye(m); eye(m) -EF']);
   temp3 = [CF zeros(m, n); zeros(m,n) BF']; 
   H = [A zeros(n); zeros(n) -A'] + temp1*temp2*temp3;
   % Compute eigenvalues and eigenvectors
   [V,L]=eig(H);  
   Vinv=inv(V');

   % Compute candidate frequency interval end-points
   nw=0;
   wtil = [];
   for i=1:2*n,
      if abs(real(L(i,i))) < 20*n*sqrt(eps)*max([1,10*abs(imag(L(i,i)))])
         if imag(L(i,i)) > 0 | ( imag(L(i,i)) == 0 & L(i,i) > 0 ),
            nw=nw+1;
            wtil(nw)=imag(L(i,i));
            vhat(1:2*n,nw)=V(:,i);      % corresponding eigenvector
            vtilde(1:2*n,nw)=Vinv(:,i); % corresponding left eigenvector
         end
      end
   end

   % Compute the derivative of the Hamiltonian
   dBF = -0.5*((1/sqrt(xi))^3)*B*Dinv*temp0;
   dCF = -0.5*((1/sqrt(xi))^3)*Dk*C;
   dEF = -((1/xi)^2)*(Dk*E*Dinv)*temp0;
   dtemp1 = [dBF zeros(n, m); zeros(n, m) -dCF'];
   dtemp2 = -temp2*[ -dEF zeros(m); zeros(m) -dEF']*temp2;
   dtemp3 = [dCF zeros(m, n); zeros(m,n) dBF'];
   DH= dtemp1*temp2*temp3 + temp1*dtemp2*temp3 + temp1*temp2*dtemp3;
   clear temp1; clear temp2; clear temp3;
   clear dtemp1; clear dtemp2; clear dtemp3;
   % Compute slopes of singular value plots at the frequencies
   % where they take the value muhat 
   [wtil,I]=sort(wtil(1:nw));
   for i=1:nw,
      slope(i)=real(inv(-j*vtilde(1:2*n,I(i))'*DH*vhat(1:2*n,I(i))));
   end
   % Possible problem: This assumes the corresponding Hamiltonian
   %   eigenvalue has multiplicity one!!
   
   % Compute intervals - Note that the method implemented here is not
   % precisely the same as that described in the paper.  The method here
   % allows for infinite intervals by starting at the trial frequency
   % and first sweeping left and then sweeping right. The case where
   % the trial frequency is one of the intersection points must be handled
   % separately from the case where it is not.
   %
   % First, check if wk is one of our points
   wk_pos=0;  % The position where the trial frequency is, or would be,
	          % in the list of sigma_max intersection frequencies
   for i=1:nw,
      if abs(wtil(i)-wk) < sqrt(sqrt(eps))
         wk_pos=i;
         if (record_slope==TRUE & check_done==FALSE)
            backup(backup_ptr,1)=wk;
            backup(backup_ptr,2)=mu_k;
            backup(backup_ptr,3)=slope(i);
            if (backup_ptr ==1 )
               backup_ptr=2;
            else
               backup_ptr=1;
            end
         end
         break
      end
   end
   % Assign each frequency point to its corresponding curve
   if wk_pos == 0
      % wk was not a frequency point
      % Determine where wk would be in list
      for i=1:nw,
         if wtil(i)>wk
            break
         end
         wk_pos=wk_pos+1;
      end
      if wk_pos < nw
	    curve(wk_pos+1)=1;
        if wk_pos < nw-1
	       % sweep left to right
	       for i=(wk_pos+2):nw,
	          if slope(i-1)<0
	             if slope(i)<0
	   	            curve(i)=curve(i-1)+1;
	             else
	      	        curve(i)=curve(i-1);
	             end
	          else
	             if slope(i)<0
		            curve(i)=curve(i-1);
	             else
		            curve(i)=curve(i-1)-1;
	             end
	          end
	       end
	     end
      end
      if wk_pos > 0
	     curve(wk_pos)=1;
	     if wk_pos > 1
	        % sweep right to left
	        for i=(wk_pos-1):-1:1,
	           if slope(i+1)<0
	              if slope(i)<0
		             curve(i)=curve(i+1)-1;
	              else
		             curve(i)=curve(i+1);
	              end
	           else
	              if slope(i)<0
		             curve(i)=curve(i+1);
	              else
		             curve(i)=curve(i+1)+1;
	              end
	           end
	        end
	     end
      end
   else
      % wk was a frequency point
      curve(wk_pos)=1;  % definitely sigma_max
      if wk_pos < nw
         % left to right sweep
	     for i=(wk_pos+1):nw,
	        if slope(i-1)<0
	           if slope(i)<0
		          curve(i)=curve(i-1)+1;
	           else
		          curve(i)=curve(i-1);
	           end
	        else
	           if slope(i)<0
		          curve(i)=curve(i-1);
	           else
		          curve(i)=curve(i-1)-1;
	           end
	        end
	     end
      end
      if wk_pos > 1
         % right to left sweep
	     for i=(wk_pos-1):-1:1,
	        if slope(i+1)<0
	           if slope(i)<0
		          curve(i)=curve(i+1)-1;
	           else
		          curve(i)=curve(i+1);
	           end
	        else
	           if slope(i)<0
		          curve(i)=curve(i+1);
	           else
		          curve(i)=curve(i+1)+1;
	           end
	        end
	     end
      end
   end
   % Extract the frequencies corresponding to sigma_max
   nf_old=nf;
   inter_old=intervals;
   nf=0;     % "r" in the LTV paper
   for i=1:nw,
      if curve(i)==1
	     nf=nf+1;
         index(nf)=i;
      end
   end
   % Construct intervals (using slopes)
   nintold=nint;
   intervals=[];     % "Omega_k" in the LTV paper
   if nf > 0
      if slope(index(1)) < 0 & wtil(index(1))>sqrt(eps)
	     intervals(1,1)=0;
         intervals(1,2)=wtil(index(1));
         row=2; col=1; i=1;
         while (2*(row-1)+col-1) <= nf,
	        i=i+1; 
            intervals(row,col)=wtil(index(i));
            if col == 2
	           col=1;
               row=row+1;
	        else
               col=2;
	        end
	     end
	     if col == 2
            intervals(row,col)=Inf;
            nint=row;
	     else
            nint=row-1;
	     end
      elseif (slope(index(1))>0 & wtil(index(1))>sqrt(eps))
         row=1; col=1; i=0;
         while (2*(row-1)+col) <= nf,
	        i=i+1; 
            intervals(row,col)=wtil(index(i));
            if col == 2
	           col=1;
               row=row+1;
	        else
               col=2;
	        end
	     end
	     if col == 2
            intervals(row,col)=Inf;
            nint=row;
	     else
            nint=row-1;
	     end
      elseif (slope(index(1))>0 & wtil(index(1))<=sqrt(eps))
         intervals(1,1)=0;
         if nf>1
            intervals(1,2)=wtil(index(2));
         else
            intervals(1,2)=Inf;
         end
         nf=nf-1;
         row=2; col=1; i=2;
         while (2*(row-1)+col-1) <= nf,
            i=i+1; 
            intervals(row,col)=wtil(index(i));
            if col == 2
               col=1;
               row=row+1;
            else
               col=2;
            end
         end
         if col == 2
            intervals(row,col)=Inf;
            nint=row;
         else
            nint=row-1;
         end
      elseif (slope(index(1))<0 & wtil(index(1))<=sqrt(eps))
         if nf>1
            intervals(1,1)=wtil(index(2));
         else
            intervals=[];
            nint=0;
         end
         nf=nf-2;
         row=1; col=2; i=2;
         while (2*(row-1)+col-1) <= nf,
            i=i+1;
            intervals(row,col)=wtil(index(i));
            if col == 2
               col=1;
               row=row+1;
            else
               col=2;
            end
         end
         if (col == 2 & ~isempty(intervals)) | (iter == 1 & isempty(intervals))
            intervals(row,col)=Inf;
            nint=row;
         else
            nint=row-1;
         end
      elseif (iter==1 & slope(index(1))==0 & nf == 1),
         intervals(1,1)=wtil(index(1));
         intervals(1,2)=Inf;
      end
      if (iter==1 & nint == 1 & ( abs(intervals(1,2)-intervals(1,1))<sqrt(eps))),
         intervals(1,1)=wtil(index(1));
         intervals(1,2)=Inf;
      end
   end
   if isempty(intervals)
      % Need to handle this case more intelligently?
      done=TRUE;
      if info==2
         disp('No intervals found')
      end
      iwarn=1;
   end

   % If necessary, intersect with previous set of intervals
   % This portion of code follows that developed by J. Sreedhar
   % in his code for computing the real stability radius.
   if nf_old > 0 & ~isempty(intervals)
      new_int=[];
      index2=1;

      for index1=1:nint,
         while (inter_old(index2,2) < intervals(index1,1)),
            index2=index2+1;
            if index2 > nintold
	           break
            end
         end
         if index2 > nintold
            break
         end
         intersected=FALSE;
         while (inter_old(index2,1) < intervals(index1,2)),
            intersected=TRUE;
            left=max(intervals(index1,1),inter_old(index2,1));
            right=min(intervals(index1,2),inter_old(index2,2));
            new_int=[new_int; left right];
            index2=index2+1;
            if index2 > nintold
	           break
            end
         end
         if intersected == TRUE
            index2=index2-1;
         end
      end   
      if isempty(new_int)
         done=TRUE;
         if info==2
            disp('Empty intersection')
         end
         iwarn=2;
      else
         nint=length(new_int(:,1));
         clear intervals
         intervals=new_int;
      end
   end

   if info>0
      iter
      muhat
      if info>1
         wk
         w_hat
         disp('Pause -- Hit any key to continue')
         pause
      end
   end

   % Terminate if the remaining list of intervals is empty
   if (size(intervals)==0)
      done = TRUE;
      break;
   end

   if (done==TRUE)
      break
   elseif (check_done==TRUE)
      nint=nintold;
      intervals=inter_old;
   end

   %  Step 3 - Compute new wk candidates
   %           midpoint and cubic/rational interpolation
   %  Note: to allow for infinite intervals, the midpoint rule
   %  is replaced with a "nonlinear midpoint" rule: tan((atan(a)+atan(b))/2)
   
   % Find max length interval 
   scale = max(1e-5,10*max(abs(eigA)));
   maxlen=atan(intervals(1,2)/scale)-atan(intervals(1,1)/scale);
   maxind=1;
   if nint>1
      for i=2:nint
         len=atan(intervals(i,2)/scale)-atan(intervals(i,1)/scale);
         if len>maxlen
   	        maxind=i;
	        maxlen=len;
	     end
      end
   end
   wk= scale*tan( (atan(intervals(maxind,1)/scale)+atan(intervals(maxind,2)/scale))/2 );

   % Try a cubic/rational fit if there is only one interval and
   % we have enough data
   %
   % For i=1,2
   %   backup(i,1) = stored frequency
   %   backup(i,2) = stored function value (muhat curve)
   %   backup(i,3) = stored derivative
   wc=-1;
   if nint==1
      if backup(2,1)>-1
	     if (interp==1)
            % Cubic interpolation -
            %
            % Fit the function                     
            %         a*x^3 + b*x^2 + c*x + d
            % by solving for the 4 coefficients a,b,c,d
            % a = coeff(1,1), b = coeff(2,1), c = coeff(3,1), d = coeff(4,1)
            %
            M=[backup(1,1)^3, backup(1,1)^2, backup(1,1), 1;
               backup(2,1)^3, backup(2,1)^2, backup(2,1), 1;
               3*backup(1,1)^2, 2*backup(1,1), 1, 0;
               3*backup(2,1)^2, 2*backup(2,1), 1, 0];
            coeff=inv(M)*[backup(1,2); backup(2,2); 
                          backup(1,3); backup(2,3)];
            % Check if any critical points exist
            disc=4*coeff(2,1)^2 -12*coeff(1,1)*coeff(3,1);	
            if disc >= 0
               if disc < eps
	              % Only one critical point
                  wc_temp=-2*coeff(2,1)/(6*coeff(1,1));
                  % Within bounds?
                  if (wc_temp>intervals(1,1)) & (wc_temp<intervals(1,2))
                     % Check curvature
                     if (6*coeff(1,1)*wc_temp+ 2*coeff(2,1) < 0)
		                wc=wc_temp;
                     end
                  end
	           else
                  % Two critical points
                  wc_temp=(-2*coeff(2,1)-sqrt(disc))/(6*coeff(1,1));
                  % Check curvature
                  if (6*coeff(1,1)*wc_temp+ 2*coeff(2,1) > 0)
 		             wc_temp=(-2*coeff(2,1)+sqrt(disc))/(6*coeff(1,1));
                  end
	              % Within bounds?
                  if (wc_temp>intervals(1,1)) & (wc_temp<intervals(1,2))
	                 wc=wc_temp;
	              end
               end
            end
	     elseif (interp==2)
            % Rational interpolation - 
            %
            % Fit the function                     
            %         a*x^2 + b*x + c
            %         ---------------
            %            d*x + 1
            % by solving for the 4 coefficients a,b,c,d
            % a = coeff(1,1), b = coeff(2,1), c = coeff(3,1), d = coeff(4,1)
            %
            M=[backup(1,1)^2, backup(1,1), 1, -backup(1,1)*backup(1,2);
               backup(2,1)^2, backup(2,1), 1, -backup(2,1)*backup(2,2);
             2*backup(1,1), 1, 0, -backup(1,3)*backup(1,1)-backup(1,2);
             2*backup(2,1), 1, 0, -backup(2,3)*backup(2,1)-backup(2,2)];
            coeff=-inv(M)*[backup(1,2); backup(2,2); 
                           backup(1,3); backup(2,3)];
            % Compute critical point
            wc=-1;
	        if abs(coeff(4,1))> sqrt(eps)
               alpha=coeff(1,1)/coeff(4,1);
               gamma=coeff(3,1)/coeff(4,1) + coeff(2,1)/(coeff(4,1)^2);
               gamma=gamma+ coeff(1,1)/(coeff(4,1)^3);
               delta=1/coeff(4,1);
               if (abs(alpha) > sqrt(eps)) & (abs(gamma) > sqrt(eps))
                  if gamma < 0
                     wc_temp=delta+sqrt(gamma/alpha);
	              else
                     wc_temp=delta-sqrt(gamma/alpha);
                  end
                  % Within bounds?
                  if (wc_temp>intervals(1,1)) & (wc_temp<intervals(1,2))
		             wc=wc_temp;
                  end
               end
            end
         end
      end
   end

   iter = iter+1;
   mu_k_old=mu_k;
end
 
mu_nm=muhat;
if iter > 1 & iwarn == 0, iter = iter-1; end
if discrete == TRUE
   w_hat = 2*atan(w_hat);
end
